package bug;

import java.awt.Font;

import javax.swing.JFrame;

import visad.ConstantMap;
import visad.DataReference;
import visad.DataReferenceImpl;
import visad.DataRenderer;
import visad.Display;
import visad.FlatField;
import visad.FunctionType;
import visad.GraphicsModeControl;
import visad.Integer1DSet;
import visad.Real;
import visad.RealType;
import visad.Scalar;
import visad.ScalarMap;
import visad.Text;
import visad.TextControl;
import visad.TextType;
import visad.Tuple;
import visad.TextControl.Justification;
import visad.java3d.DefaultRendererJ3D;
import visad.java3d.DisplayImplJ3D;
import visad.java3d.MouseBehaviorJ3D;
import visad.java3d.ProjectionControlJ3D;

/**
 * This program will reproduce a bug in both or one of Java3D and VisAD.
 * <br />
 * The normal case for this program is to display a frame that contains a line
 * and text element. The general structure of the program is based on the demonstration 
 * provided in: http://www.ssec.wisc.edu/~billh/tutorial/s1/Section1.html
 * <br />
 * The bug case for this program is that the line or text disappears from the display.
 * <br />
 * Test System: Dual core desktop, Linux (Fedora 12), Java Version: 1.5.0_14, (and openjdk 1.6.0_17),
 * Java3D Version: 1.5.1, VISAD Version: Wed Feb 24 09:48:08 CST 2010 (and nightly for 20100418). 
 * <br />
 * Bug demonstrated on Windows XP, Java 1.6, Java3D 1.5.1.
 * <br />
 * The geometry remains in the scene graph. The occurrence is intermittent: 1/20 to 1/50 runs.
 * Symptoms suggest a race condition
 */
public class ReproduceVISADBug
{
	private void runBug() throws Exception {
		Display display = new DisplayImplJ3D("display1");
		// projection
		GraphicsModeControl dispGMC = ((DisplayImplJ3D)display).getGraphicsModeControl();
		dispGMC.setProjectionPolicy(DisplayImplJ3D.PARALLEL_PROJECTION);	
		// frame
		JFrame jframe = new JFrame("My first VisAD application");
		jframe.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		jframe.getContentPane().add(((DisplayImplJ3D)display).getComponent());
		jframe.setSize(640, 480);
		jframe.setVisible(true);
		// line
		DataReference line = makeLine(display);
		// text
		DataReference text = makeTextData(display);
		// wait
		Thread.sleep(500); // visible, wait for init to end before running		
		// same thread
		addRemoveText(display, text); // text
		Thread.sleep(1000);
		addRemoveLine(display, line); // line	
		// done, the line/text may or may not be visible
		// print scene graph, or whatever.
	}

	private void addRemoveText(Display display, DataReference dataReference) throws Exception{
		// remove
		display.removeReference(dataReference);
		// add
		ConstantMap[] textColour = {new ConstantMap(1.0, Display.Red),
				new ConstantMap(1.0, Display.Green),new ConstantMap(1.0, Display.Blue),
				new ConstantMap(1.0, Display.ZAxis)};
		DataRenderer dataRenderer = new DefaultRendererJ3D();		
		display.addReferences(dataRenderer, dataReference, textColour);
	}

	private void addRemoveLine(Display display, DataReference dataReference) throws Exception {
		// remove
		display.removeReference(dataReference);
		// add
		int thickness = 4;
		ConstantMap[] textColour = {new ConstantMap(1.0, Display.Red),
				new ConstantMap(1.0, Display.Green),new ConstantMap(1.0, Display.Blue),
				new ConstantMap(thickness, Display.LineWidth),new ConstantMap(1.0, Display.ZAxis)};
		DataRenderer dataRenderer = new DefaultRendererJ3D();		
		display.addReferences(dataRenderer, dataReference, textColour);
	}
	
	private DataReference makeLine(Display display) throws Exception{
		RealType time = RealType.getRealType("time0");
		RealType height = RealType.getRealType("height0");		
		ScalarMap timeMap = new ScalarMap(time, Display.XAxis);
		ScalarMap heightMap = new ScalarMap(height, Display.YAxis);
		FunctionType func_time_height = new FunctionType(time, height);
		display.addMap(timeMap);
		display.addMap(heightMap);				
		float[][] h_vals = new float[][]{{0.0f, 33.75f, 45.0f, 33.75f, 0.0f,} };	
		Integer1DSet time_set = new Integer1DSet(time, h_vals[0].length);
		FlatField vals_ff1  = new FlatField(func_time_height, time_set);
		vals_ff1.setSamples(h_vals);		
		DataReferenceImpl data = new DataReferenceImpl("data_ref0");
		data.setData(vals_ff1);
		int thickness = 4;
		ConstantMap[] textColour = {new ConstantMap(1.0, Display.Red),
				new ConstantMap(1.0, Display.Green),new ConstantMap(1.0, Display.Blue),
				new ConstantMap(thickness, Display.LineWidth),new ConstantMap(1.0, Display.ZAxis)};
		DataRenderer dataRenderer = new DefaultRendererJ3D();		
		display.addReferences(dataRenderer, data, textColour);		
		return data;
	}
	
	private DataReference makeTextData(Display display) throws Exception{
		TextType textType = TextType.getTextType("TimeAnnotation0");
		ScalarMap textMap = new ScalarMap(textType, Display.Text);
		display.addMap(textMap);
		TextType trackText = (TextType)textMap.getScalar();		
		float x = 15f; 
		float y = 15f;
		String text = "0000000000000000";
		Scalar[] dataValues = {new Real(RealType.Latitude,x),new Real(RealType.Longitude,y),
				new Text(trackText, text)};		
		Tuple tuple = new Tuple(dataValues);
		DataReferenceImpl data = new DataReferenceImpl("trackLabel0");
		data.setData(tuple);
		ConstantMap[] textColour = {new ConstantMap(1.0, Display.Red),
				new ConstantMap(1.0, Display.Green),new ConstantMap(1.0, Display.Blue),
				new ConstantMap(1.0, Display.ZAxis)};
		DataRenderer dataRenderer = new DefaultRendererJ3D();		
		display.addReferences(dataRenderer, data, textColour);
		// add properties to text
		// this is optional, but increases the probability of the bug occurring
		// also makes line disappear instead of the text
		addTextProperties(display, textMap);
		return data;
	}
	
	private void addTextProperties(Display display, ScalarMap textMap) throws Exception {
		int size = 6;
		Font newFont = new Font("Serif", Font.PLAIN, size);
		double[] rot = new double[3];
		double[] scale = new double[1];
		double[] trans = new double[3];
		double baseFontSize = 14;		
		ProjectionControlJ3D pcontrol = (ProjectionControlJ3D) ((DisplayImplJ3D) display).getProjectionControl();
		TextControl tcontrol = (TextControl)textMap.getControl();
		MouseBehaviorJ3D.unmake_matrix(rot, scale,trans, pcontrol.getMatrix());			
		tcontrol.setSize((size / baseFontSize) / scale[0]);
		tcontrol.setFont(newFont);
		tcontrol.setRotation(180);
		tcontrol.setJustification(Justification.LEFT);
		tcontrol.setVerticalJustification(Justification.CENTER);		
	}

	public static void main(String[] args)
	{
		try{
			new ReproduceVISADBug().runBug();
		}catch(Exception e)
		{
			System.err.println("App Exception");
			e.printStackTrace();
		}
	}
}